#include "itkCurvatureAnisotropicDiffusionImageFilter.h"
#include "itkGradientMagnitudeRecursiveGaussianImageFilter.h"
#include "itkSigmoidImageFilter.h"
#include "itkImage.h"
#include "itkFastMarchingImageFilter.h"
#include "itkBinaryThresholdImageFilter.h"
#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "itkRescaleIntensityImageFilter.h"

#include "itkImageSeriesReader.h"
#include "itkDICOMImageIO2.h"
#include "itkDICOMSeriesFileNames.h"

#include "itkImageToVTKImageFilter.h"

#include "vtkImageViewer.h"
#include "vtkRenderWindowInteractor.h"
#include "vtkMarchingCubes.h"
#include "vtkPolyDataMapper.h"
#include "vtkActor.h"
#include "vtkProperty.h"
#include "vtkRenderer.h"

int main( int argc, char *argv[] )
{
  if( argc < 11 ){
    std::cerr << "Missing Parameters " << std::endl;
    std::cerr << "Usage: " << argv[0];
    std::cerr << " inputImage  outputImage seedX seedY seedZ";
    std::cerr << " Sigma SigmoidAlpha SigmoidBeta TimeThreshold StoppingValue" << std::endl;
	system("pause");
    return 1;
  }

  typedef   float           InternalPixelType;
  const     unsigned int    Dimension = 3;
  typedef itk::Image< InternalPixelType, Dimension >  InternalImageType;

  typedef unsigned char OutputPixelType;
  typedef itk::Image< OutputPixelType, Dimension > OutputImageType;

  typedef itk::BinaryThresholdImageFilter< InternalImageType, 
                        OutputImageType    >    ThresholdingFilterType;
  ThresholdingFilterType::Pointer thresholder = ThresholdingFilterType::New();

  const InternalPixelType  timeThreshold = atof( argv[9] );
  
  thresholder->SetLowerThreshold(           0.0  );
  thresholder->SetUpperThreshold( timeThreshold  );

  thresholder->SetOutsideValue(  0  );
  thresholder->SetInsideValue(  255 );

//--> DICOM IMAGE READING

  typedef itk::ImageSeriesReader< InternalImageType >     ReaderType;

  itk::DICOMImageIO2::Pointer dicomIO = itk::DICOMImageIO2::New();

  itk::DICOMSeriesFileNames::Pointer nameGenerator = itk::DICOMSeriesFileNames::New();
  nameGenerator->SetDirectory( argv[1] );

  ReaderType::Pointer reader = ReaderType::New();

  try{
    typedef std::vector<std::string> seriesIdContainer;
    const seriesIdContainer & seriesUID = nameGenerator->GetSeriesUIDs();

    seriesIdContainer::const_iterator seriesItr = seriesUID.begin();
    seriesIdContainer::const_iterator seriesEnd = seriesUID.end();

    typedef std::vector<std::string> fileNamesContainer;
    fileNamesContainer fileNames;
	fileNames = nameGenerator->GetFileNames();

    reader->SetFileNames( fileNames );
    reader->SetImageIO( dicomIO );

    try{
      reader->Update();
	}
    catch (itk::ExceptionObject &ex){
      std::cout << ex << std::endl;
      return EXIT_FAILURE;
	}
  }
  catch (itk::ExceptionObject &ex){
    std::cout << ex;
    return EXIT_FAILURE;
  }

//<--

  typedef  itk::ImageFileWriter<  OutputImageType  > WriterType;
  WriterType::Pointer writer = WriterType::New();
  writer->SetFileName( argv[2] );

  typedef itk::RescaleIntensityImageFilter< 
                               InternalImageType, 
                               OutputImageType >   CastFilterType;

  typedef   itk::CurvatureAnisotropicDiffusionImageFilter< 
                               InternalImageType, 
                               InternalImageType >  SmoothingFilterType;

  SmoothingFilterType::Pointer smoothing = SmoothingFilterType::New();

  typedef   itk::GradientMagnitudeRecursiveGaussianImageFilter< 
                               InternalImageType, 
                               InternalImageType >  GradientFilterType;

  typedef   itk::SigmoidImageFilter<                               
                               InternalImageType, 
                               InternalImageType >  SigmoidFilterType;

  GradientFilterType::Pointer  gradientMagnitude = GradientFilterType::New();
  SigmoidFilterType::Pointer sigmoid = SigmoidFilterType::New();

  sigmoid->SetOutputMinimum(  0.0  );
  sigmoid->SetOutputMaximum(  1.0  );

//--> BINARY THRESHOLD INSTEAD OF SIGMOID

  typedef itk::BinaryThresholdImageFilter< InternalImageType, 
                        InternalImageType    >    InternalThresholdingFilterType;
  InternalThresholdingFilterType::Pointer internalThresholder = InternalThresholdingFilterType::New();

  internalThresholder->SetLowerThreshold( 0.0 );
  internalThresholder->SetUpperThreshold( 40  );

  internalThresholder->SetOutsideValue(  0  );
  internalThresholder->SetInsideValue(  255 );

//<--

  typedef  itk::FastMarchingImageFilter< InternalImageType, 
                              InternalImageType >    FastMarchingFilterType;

  FastMarchingFilterType::Pointer  fastMarching = FastMarchingFilterType::New();

  //smoothing->SetInput( reader->GetOutput() );
  //gradientMagnitude->SetInput( smoothing->GetOutput() );
  gradientMagnitude->SetInput( reader->GetOutput() );
  //sigmoid->SetInput( gradientMagnitude->GetOutput() );
  internalThresholder->SetInput( gradientMagnitude->GetOutput() );
  fastMarching->SetInput( internalThresholder->GetOutput() );
  thresholder->SetInput( fastMarching->GetOutput() );
  writer->SetInput( thresholder->GetOutput() );

  smoothing->SetTimeStep( 0.0625 );
  smoothing->SetNumberOfIterations(  5 );
  smoothing->SetConductanceParameter( 9.0 );

  const double sigma = atof( argv[6] );

  gradientMagnitude->SetSigma(  sigma  );

  const double alpha =  atof( argv[7] );
  const double beta  =  atof( argv[8] );

  sigmoid->SetAlpha( alpha );
  sigmoid->SetBeta(  beta  );

  typedef FastMarchingFilterType::NodeContainer           NodeContainer;
  typedef FastMarchingFilterType::NodeType                NodeType;
  NodeContainer::Pointer seeds = NodeContainer::New();  

  InternalImageType::IndexType  seedPosition;
  
  seedPosition[0] = atoi( argv[3] );
  seedPosition[1] = atoi( argv[4] );
  seedPosition[2] = atoi( argv[5] );

  NodeType node;
  const double seedValue = 0.0;
  
  node.SetValue( seedValue );
  node.SetIndex( seedPosition );

  seeds->Initialize();
  seeds->InsertElement( 0, node );

  fastMarching->SetTrialPoints(  seeds  );

  fastMarching->SetOutputSize( 
           reader->GetOutput()->GetBufferedRegion().GetSize() );
  const double stoppingTime = atof( argv[10] );

  fastMarching->SetStoppingValue(  stoppingTime  );

  try
    {
    writer->Update();
	printf("output\n");
    }
  catch( itk::ExceptionObject & excep )
    {
    std::cerr << "Exception caught !" << std::endl;
    std::cerr << excep << std::endl;
    }

  printf("fastMarching\n");

  typedef itk::ImageToVTKImageFilter<OutputImageType> ConnectorType;

  ConnectorType::Pointer thresholdConnector= ConnectorType::New();
  thresholdConnector->SetInput( thresholder->GetOutput() );

  vtkImageViewer* viewer= vtkImageViewer::New();
  vtkRenderWindowInteractor* renderWindowInteractor = vtkRenderWindowInteractor::New();

  viewer->SetupInteractor( renderWindowInteractor);
  viewer->SetInput( thresholdConnector->GetOutput() );
  viewer->Render();
  viewer->SetColorWindow( 255);
  viewer->SetColorLevel( 128);

  typedef itk::ImageToVTKImageFilter< InternalImageType > ConnectorType2;
  
  ConnectorType2::Pointer readerConnector= ConnectorType2::New();
  readerConnector->SetInput( reader->GetOutput() );

  vtkImageViewer* viewer2= vtkImageViewer::New();
  vtkRenderWindowInteractor* renderWindowInteractor2 = vtkRenderWindowInteractor::New();

  viewer2->SetupInteractor( renderWindowInteractor2);
  viewer2->SetInput( readerConnector->GetOutput() );
  viewer2->Render();
  viewer2->SetColorWindow( 255);
  viewer2->SetColorLevel( 128);

//--> VISUALIZATION USING VTK

  vtkMarchingCubes *isoMC = vtkMarchingCubes::New();
  isoMC->SetInput(thresholdConnector->GetOutput());
  isoMC->SetValue(0, 128);

  vtkPolyDataMapper *isoMapperMC = vtkPolyDataMapper::New();
  isoMapperMC->SetInput(isoMC->GetOutput());
  isoMapperMC->ScalarVisibilityOff();
  isoMapperMC->ImmediateModeRenderingOn();

  vtkActor *isoActorMC = vtkActor::New();
  isoActorMC->SetMapper(isoMapperMC);
  isoActorMC->GetProperty()->SetColor( 0, 1, 0);
  isoActorMC->GetProperty()->SetOpacity(0.5);


  vtkRenderer *ren = vtkRenderer::New();
  vtkRenderWindow *renWin = vtkRenderWindow::New();
  renWin->AddRenderer(ren);
  vtkRenderWindowInteractor *iren = vtkRenderWindowInteractor::New();
  iren->SetRenderWindow(renWin);

  ren->AddActor(isoActorMC);

  ren->SetBackground(1, 1, 1);
  renWin->SetSize(700, 500);  
  renWin->SetPosition(50, 50);

  try{
	  iren->Initialize();
	  iren->Start();
  }
  catch( itk::ExceptionObject & excep ){
	  std::cerr << "Exception caught !" << std::endl;
	  std::cerr << excep << std::endl;
  }
  
  renderWindowInteractor->Start();
  renderWindowInteractor2->Start();

//<--

  return 0;
}
